from scipy import sparse
import numpy as np
import pyssht
from warnings import warn


class Measurement:
    """
    Base class.

    :param int ndata: number of observed data points
    :param int npix: number of pixels in image
    """

    def __init__(self, ndata, npix):
        self.ndata = ndata
        self.npix = npix

    def forward(self, X):
        """
        Forward modelling from image to observations.  Implemented by user in custom child class.

        :param X: image
        :return: observations
        """
        raise NotImplementedError

    def adjoint(self, Y):
        """
        Adjoint modelling from observations to image.  Not necessarily the inverse of :meth:`self.forward`.  Implemented by user in custom child class.

        :param Y: observations
        :return: image
        """
        raise NotImplementedError


class Identity(Measurement):
    """
    Identity measurement operator i.e. what goes in comes out.
    """

    def __init__(self, ndata, npix):
        super().__init__(ndata, npix)
        self.eye = sparse.eye(self.ndata, self.npix)
        self.eye_adj = self.eye.getH()

    def forward(self, X):
        """:meta private:"""
        assert len(X) == self.npix
        return self.eye.dot(X)

    def adjoint(self, Y):
        """:meta private:"""
        assert len(Y) == self.ndata
        return self.eye_adj.dot(Y)


class PathIntegral(Measurement):
    """
    Path integration using a matrix that describes a set of paths.

    .. todo::
       Since this is just a matrix multiplication, can be renamed to something more generic.

    :param path_matrix: :math:`N_{\mathrm{paths}}\\times N_{\mathrm{pix}}` matrix describing a set of paths.
    """

    def __init__(self, path_matrix):
        self.path_matrix = path_matrix
        self.path_matrix_adj = self.path_matrix.getH()

        self.ndata, self.npix = path_matrix.shape

    def forward(self, X):
        """:meta private:"""
        assert len(X) == self.npix
        return self.path_matrix.dot(X)

    def adjoint(self, Y):
        """:meta private:"""
        assert len(Y) == self.ndata
        return self.path_matrix_adj.dot(Y)


class WeakLensingHarmonic(Measurement):
    """
    Weak Gravitational Lensing spherical Forward model in spherical harmonic space
    """

    def __init__(self, L, mask=None, ngal=None):
        """Construct class to hold the spherical forward and
        forward adjoint operators.

        Args:

                L (int): Spherical harmonic bandlimit
                mask (int array): Map of realspace masking.
                ngal (int array): Map of galaxy observation count.

        Raises:

                ValueError: Raised if L is not positive
                ValueError: Raised if mask is the wrong shape.
                WarningLog: Raised if L is very large.
        """
        if L < 1:
            raise ValueError("Bandlimit {} must be greater than 0.".format(L))

        if L > 1024:
            warn("Bandlimit {} is very large, computational price is large.".format(L))

        # General class members
        self.L = L
        self.shape = (self.L ** 2,)

        # Define harmonic transforms and kernel mapping
        self.harmonic_kernel = self.compute_harmonic_kernel()

        # Intrinsic ellipticity dispersion
        self.var_e = 0.37 ** 2

    def forward(self, klm):
        """Spherical weak lensing measurement operator

        Args:

                klm (complex array): Convergence signal in harmonic space
        """
        # Map to shear harmonic coefficients
        return self.harmonic_mapping(klm)

    def adjoint(self, glm):
        """Spherical weak lensing adjoint measurement operator

        Args:

                glm (complex array): Shear Observations in harmonic space
        """
        return self.harmonic_mapping(glm)

    def sks_estimate(self, glm):
        """Computes spherical Kaiser-Squires estimator (for first estimate)

        Args:

                glm (complex array): Shear Observations in harmonic space
        """
        return self.harmonic_inverse_mapping(glm)

    def compute_harmonic_kernel(self):
        """Compuptes harmonic space kernel mapping."""
        k = np.ones(self.L ** 2, dtype=float)
        index = 4
        for l in range(2, self.L):
            for m in range(-l, l + 1):
                el = float(l)
                k[index] = -1.0 * np.sqrt(((el + 2.0) * (el - 1.0)) / ((el + 1.0) * el))
                index += 1
        return k

    def harmonic_mapping(self, flm):
        """Applys harmonic space mapping.

        Args:
                flm (complex array): harmonic coefficients.

        """
        out = flm * self.harmonic_kernel
        out[:4] = 0
        return out

    def harmonic_inverse_mapping(self, flm):
        """Applys harmonic space inverse mapping.

        Args:
                flm (complex array): harmonic coefficients.

        """
        out = flm / self.harmonic_kernel
        out[:4] = 0
        return out


class WeakLensing(WeakLensingHarmonic):
    """
    Weak Gravitational Lensing spherical Forward model in pixel space
    """
    def __init__(self, L, mask=None, ngal=None):
        super().__init__(L, mask, ngal)

        self.shape = (self.L, 2 * self.L - 1)

        # Define realspace masking
        if mask is None:
            self.mask = np.ones(self.shape, dtype=bool)
        else:
            self.mask = mask.astype(bool)

        # Define observational covariance
        if ngal is None:
            self.inv_cov = self.mask_forward(np.ones(self.shape))
        else:
            self.inv_cov = self.ngal_to_inv_cov(ngal)

        if self.mask.shape != self.shape:
            raise ValueError("Shape of mask map is incorrect!")

    def forward(self, kappa):
        """Spherical weak lensing measurement operator

        Args:

                kappa (complex array): Convergence signal
        """
        return self._forward(kappa, masking=True, cov_weighting=True)

    def adjoint(self, gamma):
        return self._adjoint(gamma, masking=True, cov_weighting=True)

    def _forward(self, kappa, masking=False, cov_weighting=False):
        kappa = kappa.reshape(self.shape)
        klm = pyssht.forward(kappa, self.L, Spin=0)
        glm = super().forward(klm)
        gamma = pyssht.inverse(glm, self.L, Spin=2)
        if masking:
            gamma = self.mask_forward(gamma)
        if cov_weighting:
            gamma = self.cov_weight(gamma)
        return gamma.flatten()

    def _adjoint(self, gamma, masking=False, cov_weighting=False):
        if cov_weighting:
            gamma = self.cov_weight(gamma)
        if masking:
            gamma = self.mask_adjoint(gamma)
        glm = pyssht.inverse_adjoint(gamma, self.L, Spin=2)
        klm = super().adjoint(glm)
        kappa = pyssht.forward_adjoint(klm, self.L, Spin=0)
        return kappa.flatten()

    def mask_forward(self, f):
        """Applies given mask to a field.

        Args:

                f (complex array): Realspace Signal

        Raises:

                ValueError: Raised if signal is nan
                ValueError: Raised if signal is of incorrect shape.

        """
        if f is not f:
            raise ValueError("Signal is NaN.")

        if f.shape != self.shape:
            raise ValueError("Signal shape is incorrect for mw-sampling")

        return f[self.mask]

    def mask_adjoint(self, x):
        """Applies given mask adjoint to observations

        Args:

                x (complex array): Set of observations.

        Raises:

                ValueError: Raised if signal is nan

        """
        if x is not x:
            raise ValueError("Signal is NaN.")

        f = np.zeros(self.shape, dtype=complex)
        f[self.mask] = x
        return f

    def ngal_to_inv_cov(self, ngal):
        """Converts galaxy number density map to
        data covariance.

        Assumes no correlation between pixels.

        Args:
                ngal (real array): pixel space map of number of observations per pixel.

        """
        ngal_m = self.mask_forward(ngal)
        return np.sqrt((2.0 * ngal_m) / (self.var_e))

    def cov_weight(self, x):
        """Applies covariance weighting to observations.

        Assumes no correlation between pixels.

        Args:
                x (array): pixel space map to be inverse covariance weighted.

        """
        return x * self.inv_cov



import numpy as np
from scipy.stats import laplace
from pxmcmc.utils import chebyshev1, cheb1der


class PxMCMCParams:
    """
    Class to store tuning and runtime parameters

    :param lmda: prox parameter. tuned to make proxf abritrarily close to f
    :param delta: Forward-Euler approximation step-size (MYULA, PxMALA)
    :param mu: regularization parameter
    :param s: max order of Chebyshev polynomials (SKROCK)
    :param nsamples: number of desired samples to be saved
    :param nburn: burn-in size
    :param ngap: Thinning parameter=number of iterations between saved samples
    :param complex: :code:`True` if sampled parameters are complex. Default :code:`False`.
    :param verbosity: print every :code:`verbosity` samples to console
    :param track: list of variables to keep track of
    """
    def __init__(
        self,
        lmda=3e-5,
        delta=1e-5,
        s=1,
        mu=1,
        nsamples=int(1e6),
        nburn=int(1e3),
        ngap=int(1e2),
        complex=False,
        verbosity=100,
        track=["logposterior", "L2", "prior", "chain"],
    ):
        self.lmda = lmda
        self.delta = delta
        self.mu = mu
        self.s = s
        self.nsamples = nsamples
        self.nburn = nburn
        self.ngap = ngap
        self.complex = complex
        self.verbosity = verbosity
        self.track = track


class PxMCMC:
    """
    Base class with general PxMCMC functions.
    Children of this class must implement a run function.

    :param forward: type :class:`forward.ForwardOperator` for forward modelling and gradient calculation.
    :param prior: object that implements prior and proximal calculations e.g. :class:`prior.L1`.
    :param mcmcparams: type :class:`.PxMCMCParams`
    """

    def __init__(self, forward, prior, mcmcparams=PxMCMCParams()):
        self.forward = forward
        self.prior = prior
        for attr in mcmcparams.__dict__.keys():
            setattr(self, attr, getattr(mcmcparams, attr))
        self._initialise_tracking_arrays()

    def run(self, start_point=None):
        """
        Implementation of the MCMC

        :param start_point: vector of model parameters for the start point of the chain (default :code:`None`).
        """
        raise NotImplementedError

    def logpi(self, X, preds):
        """
        Calculates the log(posterior), L2-norm and prior-norm of a model.

        :param X: vector of model parameters
        :param preds: vector of data predictions from model X
        """
        diff = self.forward.data - preds
        L2 = np.vdot(diff, self.forward.invcov @ diff)
        prior = self.prior.prior(X)
        logPi = -self.mu * prior - L2
        return logPi, L2, prior

    def _gradlogpi(self, X, preds=None):
        gradf = (X - self.prior.proxf(X)) / self.lmda
        if preds is None:
            preds = self.forward.forward(X)
        gradg = self.forward.calc_gradg(preds)
        return -gradf - gradg

    def _print_progress(self, i, logpi, **kwargs):
        print(
            f"{i+1:,}/{self.nsamples:,} - logposterior: {logpi:.8e} - "
            + " - ".join([f"{k}: {kwargs[k]:.8e}" for k in kwargs]),
        )

    def _initial_sample(self, initial_sample=None):
        if initial_sample is None:
            # TODO: flexibility for different priors
            X_curr = laplace.rvs(size=self.forward.nparams)
            if self.complex:
                X_curr = X_curr + laplace.rvs(size=self.forward.nparams) * 1j
        else:
            if not isinstance(initial_sample, np.ndarray) or np.ndim(initial_sample) != 1:
                raise TypeError("Expected a 1D numpy array as an initial sample")
            if initial_sample.size == self.forward.nparams:
                X_curr = initial_sample
            else:
                raise ValueError("Inital sample given has incorrect size")
        curr_preds = self.forward.forward(X_curr)
        return X_curr, curr_preds

    def _initialise_tracking_arrays(self):
        if "logposterior" in self.track:
            self.logPi = np.zeros(self.nsamples)
        if "predictions" in self.track:
            self.preds = np.zeros(
                (self.nsamples, len(self.forward.data)), dtype=float
            )
        if "chain" in self.track:
            self.chain = np.zeros(
                (self.nsamples, self.forward.nparams),
                dtype=complex if self.complex else float,
            )
        if "L2" in self.track:
            self.L2s = np.zeros(self.nsamples, dtype=float)
        if "prior" in self.track:
            self.priors = np.zeros(self.nsamples, dtype=float)

    def _tracking(self, j, X_curr, curr_preds, logPi, L2, prior):
        if hasattr(self, "logPi"):
            self.logPi[j] = logPi
        if hasattr(self, "L2s"):
            self.L2s[j] = L2
        if hasattr(self, "priors"):
            self.priors[j] = prior
        if hasattr(self, "preds"):
            self.preds[j] = curr_preds
        if hasattr(self, "chain"):
            self.chain[j] = X_curr


class MYULA(PxMCMC):
    """
    Implements the MYULA chain
    """
    def __init__(self, forward, prox, mcmcparams=PxMCMCParams()):
        super().__init__(forward, prox, mcmcparams)

    def run(self, start_point=None):
        """
        Run the algorithm
        """
        i = 0  # total samples
        j = 0  # saved samples (excludes burn-in and thinned samples)
        X_curr, curr_preds = self._initial_sample(start_point)
        while j < self.nsamples:
            gradg = self.forward.calc_gradg(curr_preds)
            proxf = self.prior.proxf(X_curr)
            X_prop = self.chain_step(X_curr, proxf, gradg)
            prop_preds = self.forward.forward(X_prop)

            X_curr = X_prop
            curr_preds = prop_preds

            if i >= self.nburn:
                if self.ngap == 0 or (i - self.nburn) % self.ngap == 0:
                    logPi, L2, prior = self.logpi(X_curr, curr_preds)
                    self._tracking(j, X_curr, curr_preds, logPi, L2, prior)
                    j += 1
                if self.verbosity > 0 and (i + 1) % self.verbosity == 0:
                    self._print_progress(
                        j - 1,
                        self.logPi[j - 1],
                        L2=self.L2s[j - 1],
                        prior=self.priors[j - 1],
                    )
            else:
                if self.verbosity > 0 and (i + 1) % self.verbosity == 0:
                    print("Burning in...")
            i += 1

        print("\nDONE")

    def chain_step(self, X, proxf, gradg):
        """
        Takes a step in the chain.

        :param X: current sample
        :param proxf: prox of prior probablity
        :param grardg: gradient of data fidelity
        """
        w = np.random.randn(self.forward.nparams)
        if self.complex:
            w = w + np.random.randn(self.forward.nparams) * 1j
        return (
            (1 - self.delta / self.lmda) * X
            + (self.delta / self.lmda) * proxf
            - self.delta * gradg
            + np.sqrt(2 * self.delta) * w
        )


class PxMALA(MYULA):
    """
    Implements the PxMALA algorithm (MYULA + MH acceptance)

    :param bool tune_delta: if :code:`True`, tunes  :code:`delta` parameter to achieve an acceptance probablity of 0.5
    
    .. todo::

       Option to tune :code:`delta` to any desired acceptance probability.
    """
    def __init__(self, forward, prox, mcmcparams=PxMCMCParams(), tune_delta=True):
        super().__init__(forward, prox, mcmcparams)
        self.tune_delta = tune_delta

    def run(self, start_point=None):
        """
        Run the algorithm
        """
        self.acceptance_trace = []
        self.deltas_trace = [self.delta]
        i = 0
        j = 0
        X_curr, curr_preds = self._initial_sample(start_point)
        gradg_curr = self.forward.calc_gradg(curr_preds)
        proxf_curr = self.prior.proxf(X_curr)
        logpiXc, L2Xc, priorXc = self.logpi(X_curr, curr_preds)
        while j < self.nsamples:
            X_prop = self.chain_step(X_curr, proxf_curr, gradg_curr)
            prop_preds = self.forward.forward(X_prop)
            gradg_prop = self.forward.calc_gradg(prop_preds)
            proxf_prop = self.prior.proxf(X_prop)

            logtransXcXp = self.calc_logtransition(
                X_curr, X_prop, proxf_curr, gradg_curr
            )
            logtransXpXc = self.calc_logtransition(
                X_prop, X_curr, proxf_prop, gradg_prop
            )
            logpiXp, L2Xp, priorXp = self.logpi(X_prop, prop_preds)

            logalpha = logtransXpXc + logpiXp - logtransXcXp - logpiXc
            accept = np.log(np.random.rand()) < logalpha
            if accept:
                X_curr = X_prop
                curr_preds = prop_preds
                gradg_curr = gradg_prop
                proxf_curr = proxf_prop
                logpiXc = logpiXp
                L2Xc = L2Xp
                priorXc = priorXp
                self.acceptance_trace.append(1)
            else:
                self.acceptance_trace.append(0)

            if self.tune_delta:
                self._tune_delta(i)
                self.deltas_trace.append(self.delta)

            if i >= self.nburn:
                if (self.ngap == 0 or (i - self.nburn) % self.ngap == 0) and accept:
                    self._tracking(j, X_curr, curr_preds, logpiXc, L2Xc, priorXc)
                    j += 1
            if self.verbosity > 0 and (i + 1) % self.verbosity == 0:
                self._print_progress(
                    j - 1,
                    logpiXc,
                    L2=L2Xc,
                    prior=priorXc,
                    acceptanceRate=np.mean(self.acceptance_trace),
                )
            i += 1
        print("\nDONE")

    def _tune_delta(self, i):
        delta = self.delta * (1 + (self.acceptance_trace[i] - 0.5) / ((i + 1) ** 0.75))
        self.delta = min(max(delta, self.lmda * 1e-8), self.lmda / 2)

    def calc_logtransition(self, X1, X2, proxf, gradg):
        """
        Calculates the transition probability of stepping from model X1 to model X2 i.e. q(X2|X1).
        """
        gradlogpiX1 = -((X1 - proxf) / self.lmda) - gradg
        return (
            -(1 / 2 * self.delta)
            * np.sum((X2 - X1 - (self.delta / 2) * gradlogpiX1) ** 2) ** 2
        )


class SKROCK(PxMCMC):
    """
    Implements the SKROCK algorithm

    .. warning::
    
       This algorithm hasn't been used in practice yet from this code package
    """
    def __init__(self, forward, prox, mcmcparams=PxMCMCParams()):

        super().__init__(forward, prox, mcmcparams=mcmcparams)
        self.eta = 0.05
        self.omega_0 = 1 + self.eta / (self.s * self.s)
        self.omega_1 = chebyshev1(self.omega_0, self.s) / cheb1der(self.omega_0, self.s)
        self._recursion_coefs()

    def run(self, start_point=None):
        """
        Run the algorithm
        """
        i = 0  # total samples
        j = 0  # saved samples (excludes burn-in and thinned samples)
        X_curr, curr_preds = self._initial_sample(start_point)
        while j < self.nsamples:
            X_prop = self.chain_step(X_curr)
            prop_preds = self.forward.forward(X_prop)

            X_curr = X_prop
            curr_preds = prop_preds

            if i >= self.nburn:
                if self.ngap == 0 or (i - self.nburn) % self.ngap == 0:
                    logPi, L2, prior = self.logpi(X_curr, curr_preds)
                    self._tracking(j, X_curr, curr_preds, logPi, L2, prior)
                    j += 1
            if self.verbosity > 0 and (i + 1) % self.verbosity == 0:
                self._print_progress(
                    j - 1,
                    self.logPi[j - 1],
                    L2=self.L2s[j - 1],
                    prior=self.priors[j - 1],
                )
            i += 1

        print("\nDONE")

    def chain_step(self, X):
        """
        Takes a step in the chain.

        :param X: current sample
        """
        Z = np.random.randn(len(X))
        if self.complex:
            Z = Z + np.random.randn(len(X)) * 1j
        return self._K_recursion(X, self.s, Z)

    def _K_recursion(self, X, s, Z):
        if s == 0:
            return X
        elif s == 1:
            return (
                X
                + self.mus[1]
                * self.delta
                * self._gradlogpi(X + self.nus[1] * np.sqrt(2 * self.delta) * Z)
                + self.ks[1] * np.sqrt(2 * self.delta) * Z
            )
        else:
            return (
                self.mus[s]
                * self.delta
                * self._gradlogpi(self._K_recursion(X, s - 1, Z))
                + self.nus[s] * self._K_recursion(X, s - 1, Z)
                + self.ks[s]
                - self._K_recursion(X, s - 2, Z)
            )

    def _recursion_coefs(self):
        self.mus = np.zeros(self.s + 1)
        self.nus = np.zeros(self.s + 1)
        self.ks = np.zeros(self.s + 1)

        self.mus[1] = self.omega_1 / self.omega_0
        self.nus[1] = self.s * self.omega_1 / 2
        self.ks[1] = self.s * self.omega_1 / self.omega_0

        for j in range(2, self.s + 1):
            cheb_ratio = chebyshev1(self.omega_0, j - 1) / chebyshev1(self.omega_1, j)
            self.mus[j] = 2 * self.omega_1 * cheb_ratio
            self.nus[j] = 2 * self.omega_0 * cheb_ratio
            self.ks[j] = 1 - self.nus[0]



import os
import h5py


def save_mcmc(
    mcmc, params, outpath, filename="outputs", **kwargs,
):
    """
    Saves the MCMC run in a :code:`.hd5f` file. Any variable defined by a :class:`mcmc.PxMCMC` instance's :code:`tracking` option will be saved as a `dataset <https://docs.h5py.org/en/stable/high/dataset.html>`_.  Runtime parameters and other things that the user may wish to save are saved as  `attributes <https://docs.h5py.org/en/stable/high/attr.html>`_.

    :param mcmc: type :class:`mcmc.PxMCMC` after its :meth:`run` has been called.
    :param params: type :class:`mcmc.PxMCMCParams` used to setup the :code:`mcmc` instance.
    :param string outpath: directory in which file should be saved.
    :param string filename: filename without extension.
    :param \**kwargs: additional bits users want to save.
    """
    with h5py.File(os.path.join(outpath, f"{filename}.hdf5"), "w") as f:
        if hasattr(mcmc, "logPi"):
            f.create_dataset("logposterior", data=mcmc.logPi)
        if hasattr(mcmc, "preds"):
            f.create_dataset("predictions", data=mcmc.preds)
        if hasattr(mcmc, "chain"):
            f.create_dataset("chain", data=mcmc.chain)
        if hasattr(mcmc, "L2s"):
            f.create_dataset("L2s", data=mcmc.L2s)
        if hasattr(mcmc, "priors"):
            f.create_dataset("priors", data=mcmc.priors)
        if hasattr(mcmc, "acceptance_trace"):
            f.create_dataset("acceptances", data=mcmc.acceptance_trace, dtype="i1")
        if hasattr(mcmc, "deltas_trace"):
            f.create_dataset("deltas", data=mcmc.deltas_trace)

        for attr in params.__dict__.keys():
            f.attrs[attr] = getattr(params, attr)
        for k, v in kwargs.items():
            f.attrs[k] = v



import numpy as np
import healpy as hp
import matplotlib.pyplot as plt
from matplotlib import cm
import copy
import pyssht
import pys2let

try:
    from cartopy.crs import Mollweide
except ModuleNotFoundError:
    print("cartopy not found.  Cannot plot coasts.")

from pxmcmc.utils import suppress_stdout, _multires_bandlimits


def plot_map(
    f,
    title=None,
    cbar=True,
    cmap="turbo",
    vmin=None,
    vmax=None,
    cbar_label="",
    oversample=True,
    centre0=False,
    coasts=False,
    mask=None,
):
    """
    Plots a single `MW <https://arxiv.org/abs/1110.6298>`_ sampled spherical map.

    :param array f: MW sampled image.  Shape :math:`(L, 2L - 1)`.
    :param string title: Figure title
    :param bool cbar: if :code:`True`, plot the colour bar
    :param string cmap: Name of a `matplotlib colour map <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_
    :param float vmin: Minimum value.  If :code:`None`, will default to lowest value in :code:`f`.
    :param float vmax: Maximum value.  If :code:`None`, will default to highest value in :code:`f`.
    :param string cbar_label: Label for the colour bar. Requires :code:`cbar=True`
    :param bool oversample: if :code:`True`, oversamples :code:`f` to bandlimit :math:`L=256` so the image is not pixelated
    :param bool centre0: if :code:`True`, forces the colour map to be centred at 0.  Overrides :code:`vmin,vmax`.
    :param bool coasts: if :code:`True`, plots coastlines.
    :param boolarray mask: A binary MW mask. Mask is applied after any oversampling, so should be a higher resolution than :code:`f`

    :return: matplotlib figure
    """
    cmap = copy.copy(cm.get_cmap(cmap))
    cmap.set_bad(alpha=0)

    if oversample:
        L = 256
        f = _oversample(f, L)
    else:
        L = f.shape[0]

    if centre0:
        cbar_end = max([f.max(), abs(f.min())])
        vmax = cbar_end
        vmin = -cbar_end

    if mask is not None:
        f[mask] = np.nan
    f_plt, _ = pyssht.mollweide_projection(f, L)
    fig = plt.figure(figsize=(20, 10))
    if not cbar:
        map_gs = fig.add_gridspec(1, 1)
        map_ax = map_gs.subplots()
        im = map_ax.imshow(f_plt, origin="lower", cmap=cmap, vmin=vmin, vmax=vmax)
    else:
        map_gs = fig.add_gridspec(nrows=1, ncols=2, width_ratios=[40, 1], wspace=0.05)
        map_ax = fig.add_subplot(map_gs[:, :-1])
        cbar_ax = fig.add_subplot(map_gs[:, -1])
        im = map_ax.imshow(f_plt, origin="lower", cmap=cmap, vmin=vmin, vmax=vmax)
        cbar = fig.colorbar(im, cax=cbar_ax)
        cbar.set_label(cbar_label, fontsize=24)
        cbar.ax.tick_params(labelsize="xx-large")
    map_ax.axis("off")
    map_ax.set_title(title, fontsize=24)
    if coasts:
        coast_gs = map_gs[0].subgridspec(1, 1)
        coast_ax = coast_gs.subplots(subplot_kw={"projection": Mollweide()})
        coast_ax.coastlines(linewidth=2)
        coast_ax.patch.set_alpha(0)
    return fig


def plot_wavelet_maps(f, L, B, J_min, dirs=1, spin=0, same_scale=True, **map_args):
    """
    Plots the scaling and wavelet maps of spherical map :code:`f`.  
    
    :param array f: MW sampled image.  Shape :math:`(L, 2L - 1)`.
    :param int L: angular bandlimit
    :param float B: wavelet scale parameter
    :param int J_min: minimum wavelet scale
    :param int dirs: azimuthal bandlimit for directional wavelets
    :param int spin: spin number of spherical signal
    :param bool same_scale: if :code:`True`, wavelet maps are plotted on same colour scale
    :param \**map_args: optional arguments for :meth:`plot_map`

    :return: List of figures
    """
    bls = _multires_bandlimits(L, B, J_min, dirs, spin)
    f_wav, f_scal = pys2let.analysis_px2wav(
        f.flatten().astype(complex), B, L, J_min, dirs, spin, upsample=0
    )
    figs = []
    if "title" in map_args:
        base_title = map_args["title"]
    else:
        base_title = ""
    map_args["title"] = f"{base_title} Scaling function"
    figs.append(plot_map(f_scal.reshape(pyssht.sample_shape(bls[0])), **map_args))

    if same_scale:
        map_args["vmax"] = np.max(f_wav).real

    scale_start = 0
    for i, bl in enumerate(bls[1:], 1):
        scale_length = pyssht.sample_length(bl)
        wav = f_wav[scale_start : scale_start + scale_length]
        map_args["title"] = f"{base_title} Wavelet scale {i}"
        figs.append(plot_map(wav.reshape(pyssht.sample_shape(bl)), **map_args))
        scale_start += scale_length

    return figs


def mollview(image, figsize=(10, 8), **kwargs):
    i = np.random.randint(1000)
    fig = plt.figure(
        num=i, figsize=figsize
    )  # this figure number thing is a bit hacky...
    with suppress_stdout():
        hp.mollview(image, fig=i, **kwargs)
        hp.graticule(30)
    return fig


def plot_evolution(logposteriors, L2s, L1s, figsize=(10, 8)):
    """
    Plot the evolution of the MCMC chain.

    :param logposteriors: array of log posterior probabilities of the saved MCMC samples. Plot shows the negative log posterior.
    :param L2s: array of log gaussian data fidelities (L2 error norms)
    :param L1s: array of log Laplacian priors (L1 norms)
    :param tuple figsize: Figure size

    :return: matplotlib figure
    """
    MAP_idx = np.where(logposteriors == max(logposteriors))
    fig = plt.figure(figsize=figsize)
    plt.subplot(3, 1, 1)
    plt.plot(-logposteriors)
    plt.axvline(MAP_idx, linestyle="--", c="r")
    plt.yscale("log")
    plt.ylabel("-log(posterior)")

    plt.subplot(3, 1, 2)
    plt.plot(L2s)
    plt.axvline(MAP_idx, linestyle="--", c="r")
    plt.yscale("log")
    plt.ylabel("L2")

    plt.subplot(3, 1, 3)
    plt.plot(L1s)
    plt.axvline(MAP_idx, linestyle="--", c="r")
    plt.yscale("log")
    plt.ylabel("L1")
    return fig


def plot_chain_sample(X, figsize=(10, 8)):
    """
    Plots the real and imaginary parts of an MCMC sample

    :param X: MCMC sample
    :param tuple figsize: Figure size

    :return: matplotlib figure
    """
    fig = plt.figure(figsize=figsize)
    plt.subplot(2, 1, 1)
    plt.plot(X.real)
    plt.subplot(2, 1, 2)
    plt.plot(X.imag)
    return fig


def _oversample(f, L):
    flm = pyssht.forward(f, f.shape[0], Reality=True)
    z = np.zeros(L ** 2 - f.shape[0] ** 2)
    flm = np.concatenate((flm, z))
    return pyssht.inverse(flm, L, Reality=True)



import pys2let
import pyssht
import numpy as np

from pxmcmc.utils import expand_mlm, flatten_mlm


class Transform:
    """Base class to wrap transformations."""

    def forward(self):
        """
        e.g. spherical image to spherical harmonics.  Implemented by user in custom child class.
        """
        raise NotImplementedError

    def inverse(self):
        """
        e.g. spherical harmonics to spherical image.  Implemented by user in custom child class.
        """
        raise NotImplementedError

    def forward_adjoint(self):
        """
        e.g. spherical harmonics to spherical image.  Implemented by user in custom child class.
        """
        raise NotImplementedError

    def inverse_adjoint(self):
        """
        e.g. spherical image to spherical harmonics.  Implemented by user in custom child class.
        """
        raise NotImplementedError


class IdentityTransform(Transform):
    """Identity transform i.e. does nothing"""

    def __init__(self):
        pass

    def forward(self, X):
        """:meta private:"""
        return X

    def forward_adjoint(self, X):
        """:meta private:"""
        return X

    def inverse(self, X):
        """:meta private:"""
        return X

    def inverse_adjoint(self, X):
        """:meta private:"""
        return X


class SphericalWaveletTransform(Transform):
    """
    Spherical wavelet transforms

    :param int L: angular bandlimit
    :param float B: wavelet scale parameter
    :param int J_min: minimum wavelet scale
    :param int dirs: azimuthal bandlimit for directional wavelets
    :param int spin: spin number of spherical signal
    :param bool harmonic: inputs and outputs in spherical harmonic space
    """

    def __init__(self, L, B, J_min, dirs=1, spin=0, harmonic=False):
        self.L = L
        self.B = B
        self.J_min = J_min
        self.J_max = pys2let.pys2let_j_max(self.B, self.L, self.J_min)
        self.nscales = self.J_max - self.J_min + 1
        self.dirs = dirs
        self.spin = spin

        self.params = {
            "B": B,
            "L": L,
            "J_min": J_min,
            "N": dirs,
            "spin": spin,
            "upsample": 0,
        }

        if harmonic:
            self.fwd = pys2let.analysis_lm2lmn
            self.fwd_adj = pys2let.analysis_adjoint_lmn2lm
            self.inv = pys2let.synthesis_lmn2lm
            self.inv_adj = pys2let.synthesis_adjoint_lm2lmn
        else:
            self.fwd = pys2let.analysis_px2wav
            self.fwd_adj = pys2let.analysis_adjoint_wav2px
            self.inv = pys2let.synthesis_wav2px
            self.inv_adj = pys2let.synthesis_adjoint_px2wav

        self._get_ncoefs()

    def forward(self, X):
        """
        Transform image to spherical wavelet space.

        :param X: spherical image as a 1D array
        :return: 1D array of spherical wavelet coefficients
        """
        if not isinstance(X, complex):
            X = X.astype(complex)
        X_wav, X_scal = self.fwd(X, **self.params)
        return flatten_mlm(X_wav, X_scal)

    def inverse(self, X):
        """
        Transform spherical wavelet to image space.

        :param X: 1D array of spherical wavelet coefficients
        :return: spherical image as a 1D array
        """
        wav, scal = expand_mlm(X, nscalcoefs=self.nscal)
        if not isinstance(scal, complex):
            scal = scal.astype(complex)
        if not isinstance(scal, complex):
            wav = wav.astype(complex)
        X = self.inv(wav, scal, **self.params)
        return X

    def inverse_adjoint(self, X):
        """
        Adjoint transform image to spherical wavelet space.

        :param X: spherical image as a 1D array
        :return: 1D array of spherical wavelet coefficients
        """
        if not isinstance(X, complex):
            X = X.astype(complex)
        X_wav, X_scal = self.inv_adj(X, **self.params)
        return flatten_mlm(X_wav, X_scal)

    def forward_adjoint(self, X):
        """
        Transform spherical wavelet to image space.

        :param X: 1D array of spherical wavelet coefficients
        :return: spherical image as a 1D array
        """
        wav, scal = expand_mlm(X, nscalcoefs=self.nscal)
        if not isinstance(scal, complex):
            scal = scal.astype(complex)
        if not isinstance(scal, complex):
            wav = wav.astype(complex)
        X = self.fwd_adj(wav, scal, **self.params)
        return X

    def _get_ncoefs(self):
        """
        Counts the number of wavelet and scaling coefs for
        the multiresolution algorithm

        TODO: CHECK THIS FOR HARMONIC
        """
        f_mw = np.empty(pyssht.sample_length(self.L), dtype=complex)
        f_wav, f_scal = pys2let.analysis_px2wav(f_mw, **self.params)
        self.nwav, self.nscal = (f_wav.shape[0], f_scal.shape[0])
        self.ncoefs = self.nwav + self.nscal



import numpy as np
import pys2let
import pyssht

from pxmcmc.utils import soft, mw_map_weights, _multires_bandlimits


class L1:
    """
    Base L1-norm prior.  The prox of this prior is soft thresholding.

    :param string setting: 'analysis' or 'synthesis'
    :param fwd: function handle for transform operator (e.g. :meth:`transforms.Transform.forward`)
    :param adj: function handle for adjoint transform operator (e.g. :meth:`transforms.Transform.forward_adjoint`)
    :param float T: threshold for the soft thresholding function

    .. todo::
       :code:`fwd` and :code:`adj` are only needed for analysis setting.  Make these optional arguments.
    """

    def __init__(self, setting, fwd, adj, T):
        assert setting in ["analysis", "synthesis"]
        self.setting = setting
        self.fwd = fwd
        self.adj = adj
        self.T = T

    def prior(self, X):
        """
        Calculates the logprior of mcmc sample

        :param X: MCMC sample
        :return: log prior
        """
        return sum(abs(X))

    def proxf(self, X):
        """
        Calculates the proximal map of the log prior

        :param X: MCMC sample
        :return: prox of log prior
        """
        if self.setting == "synthesis":
            return self._proxf_synthesis(X)
        else:
            return self._proxf_analysis(X)

    def _proxf_synthesis(self, X):
        return soft(X, self.T)

    def _proxf_analysis(self, X):
        return X + self.fwd(soft(self.adj(X), self.T) - self.adj(X))


class S2_Wavelets_L1(L1):
    """
    L1 regulariser for wavelets on S2 (MW sampling).  Performs some weighting to avoid overemphasizing pixels at the poles.

    :param int L: angular bandlimit
    :param float B: wavelet scale parameter
    :param int J_min: minimum wavelet scale
    :param int dirs: azimuthal bandlimit for directional wavelets
    :param int spin: spin number of spherical signal
    """

    def __init__(self, setting, fwd, adj, T, L, B, J_min, dirs=1, spin=0):
        super().__init__(setting, fwd, adj, T)
        self.L = L
        self.B = B
        self.J_min = J_min
        self.J_max = pys2let.pys2let_j_max(B, L, J_min)
        self.nscales = self.J_max - J_min + 1
        self.dirs = dirs
        self.spin = spin
        if setting == "synthesis":
            bls = _multires_bandlimits(L, B, J_min, dirs, spin)
            self.map_weights = np.concatenate([mw_map_weights(el) for el in bls])
        else:
            raise NotImplementedError
        self.T *= self.map_weights

    def prior(self, X):
        return super().prior(self.map_weights * X)


class S2_Wavelets_L1_Power_Weights(S2_Wavelets_L1):
    """
    L1 regulariser for wavelets on S2 (MW sampling).
    Includes weighting for pixel area, wavelet power wavelet decay
    See eqns 33&34 from Wallis et al 2017

    :param int L: angular bandlimit
    :param float B: wavelet scale parameter
    :param int J_min: minimum wavelet scale
    :param int dirs: azimuthal bandlimit for directional wavelets
    :param int spin: spin number of spherical signal
    :param float eta: wavelet decay tuning parameter
    """

    def __init__(self, setting, fwd, adj, T, L, B, J_min, dirs=1, spin=0, eta=1):
        super().__init__(setting, fwd, adj, T, L, B, J_min, dirs, spin)
        self.eta = eta
        if setting == "synthesis":
            self._get_weights()
        else:
            raise NotImplementedError
        self.T *= self.map_weights

    def prior(self, X):
        return super().prior(self.map_weights * X)

    def _get_weights(self):
        scaling_weights = self._calculate_scaling_weights()
        s = scaling_weights.flatten()
        wavelet_weights = self._calculate_wavelet_weights()
        w = np.concatenate([w.flatten() for w in wavelet_weights])
        self.map_weights = np.concatenate([s, w])

    def _calculate_scaling_weights(self):
        phi_l, _ = pys2let.wavelet_tiling(self.B, self.L, self.dirs, self.J_min, self.spin)
        scaling_power = np.vdot(phi_l, phi_l).real
        effective_L = np.nonzero(phi_l)[0].max() + 1
        nsamples = pyssht.sample_length(effective_L)
        weights = np.full(pyssht.sample_shape(effective_L), 2 * np.pi ** 2 / (scaling_power * nsamples))
        thetas, _ = pyssht.sample_positions(effective_L)
        weights = (weights.T * np.sin(thetas)).T
        return weights

    def _calculate_wavelet_weights(self):
        bls = _multires_bandlimits(self.L, self.B, self.J_min)
        _, psi_lm = pys2let.wavelet_tiling(self.B, self.L, self.dirs, self.J_min, self.spin)
        wavelet_powers = np.array(
            [np.vdot(lm, lm).real for lm in psi_lm.T]
        )
        psi_l = np.zeros((psi_lm.shape[1], self.L), dtype=complex)
        for j, psi in enumerate(psi_lm.T):
            psi_l[j, :] = np.array([psi[el ** 2 + el] for el in range(self.L)])
        peak_ls = np.array(
            [np.argmax(l) for l in psi_l]
        )
        all_weights = []
        for effective_L, power, peak_l in zip(bls[1:], wavelet_powers, peak_ls):
            nsamples = pyssht.sample_length(effective_L)
            weights = np.full(pyssht.sample_shape(effective_L), (2 * np.pi ** 2) * (peak_l ** self.eta) / (power * nsamples))
            thetas, _ = pyssht.sample_positions(effective_L)
            weights = (weights.T * np.sin(thetas)).T
            all_weights.append(weights)
        return np.array(all_weights, dtype=list)



import numpy as np
import pyssht

from pxmcmc.utils import _multires_bandlimits


def credible_interval_range(chain, alpha=0.05):
    """
    Calculates the range of the credible interval of MCMC samples at a given confidence level

    :param chain: :math:`N_{\mathrm{samples}}\\times N_{\mathrm{params}}` array containing MCMC samples
    :param float `alpha`: confidence level
    :return: array of credible interval ranges for each parameter
    """
    quantiles = np.quantile(chain, (alpha / 2, 1 - alpha / 2), axis=0)
    return np.diff(quantiles, axis=0)[0]


def wavelet_credible_interval_range(chain, L, B, J_min, alpha=0.05):
    """
    Maps the credible interval range at different wavelet scales. Assumes each sample in chain is a set of wavelet coefficients in multiresolution format. Maps are returned in MW (theta, phi) format.

    :param chain: :math:`N_{\mathrm{samples}}\\times N_{\mathrm{params}}` array containing MCMC samples
    :param int L: angular bandlimit
    :param float B: wavelet scale parameter
    :param int J_min: minimum wavelet scale
    :param float `alpha`: confidence level
    :return: array of credible interval ranges for each wavelet coefficient
    """
    bls = _multires_bandlimits(L, B, J_min)
    scale_start = 0
    wav_ci_ranges = []
    for i, bl in enumerate(bls):
        scale_length = pyssht.sample_length(bl)
        wav = chain[:, scale_start : scale_start + scale_length]
        wav_ci_ranges.append(
            credible_interval_range(wav, alpha).reshape(pyssht.sample_shape(bl))
        )
        scale_start += scale_length
    return wav_ci_ranges


def credible_region_threshold(logpis, alpha=0.05):
    """
    Calculates the log posterior threshold for a sample to be part of the credible set at a given confidence level.

    :param logpis: vector of log posterior probabilities of MCMC samples
    :param float `alpha`: confidence level
    :return: log posterior threshold
    """
    return np.quantile(logpis, 1 - alpha)


def in_credible_region(logpi, threshold):
    """:meta private:"""
    return True if logpi <= threshold else False



import numpy as np
import healpy as hp
from contextlib import contextmanager
import os
import sys
import pys2let
import pyssht
from astropy.coordinates import SkyCoord


def flatten_mlm(wav_lm, scal_lm):
    """
    Takes a set of wavelet and scaling coefficients and flattens them into a single vector

    :param wav_lm: array of wavelet coefficients. Gets flattened if has multiple dimensions
    :param scal_lm: array of wavelet coefficients

    :return: 1D array of coefficients, with the scaling coefficients first
    """
    buff = wav_lm.ravel(order="F")
    mlm = np.concatenate((scal_lm, buff))
    return mlm


def expand_mlm(mlm, nscales=None, nscalcoefs=None, flatten_wavs=False):
    """
    Sepatates scaling and wavelet coefficients from a single vector to separate arrays.

    :param mlm: 1D array containing scaling and wavelet coefficients.  Assumes the scaling coefficients come first followed by the wavelet coefficients at increasingly smaller scales.
    :param int nscales: number of wavlet scales + 1 scaling function. Use only if coefficients do not follow the multiresolution algorithm, otherwise leave as :code:`None`
    :param int nscalcoefs: number of scaling function coefficients.  Use only if coefficients follow the multiresolution algorithm, otherwise leave as :code:`None`
    :param bool flatten_wavs: flattens the wavelet coefficients into a 1D array, otherwise 2D.  Ignored if :code:`nscalcoefs` is not :code:`None`.

    :return: tuple (wavelet coefficients, scaling coefficients)
    """
    if nscales is None and nscalcoefs is None:
        raise ValueError("Set either 'nscales', or 'nscalcoefs'")
    elif nscales is not None and nscalcoefs is not None:
        raise ValueError("Give only one of 'nscales' or 'nscalcoefs'")
    elif nscales is not None:
        v_len = mlm.size // (nscales + 1)
        assert v_len > 0
        scal_lm = mlm[:v_len]
        wav_lm = np.zeros((v_len, nscales), dtype=complex)
        for i in range(nscales):
            wav_lm[:, i] = mlm[(i + 1) * v_len : (i + 2) * v_len]
        if flatten_wavs:
            wav_lm = np.concatenate([wav_lm[:, i] for i in range(nscales)])
    elif nscalcoefs is not None:
        scal_lm = mlm[:nscalcoefs]
        wav_lm = mlm[nscalcoefs:]
    return wav_lm, scal_lm


def soft(X, T=0.1):
    """
    Soft thresholding of a vector X with threshold T.  If :math:`X_i < T`, then :math:`\mathrm{soft}(X_i) = 0`, otherwise :math:`\mathrm{soft}(X_i) = X_i-T`.

    :param X: vector of values to threshold
    :param float T: threshold.  Can be a vector of same size as :code:`X`
    
    :return: thresholded vector
    """
    X = np.array(X)
    t = _sign(X) * (np.abs(X) - T)
    t[np.abs(X) <= T] = 0
    return t


def hard(X, T=0.1):
    """
    Hard thresholding of a vector X with fraction threshold T. T is the fraction kept, i.e. the largest 100T% absolute values are kept, the others are thresholded to 0.
    TODO: What happens when all elements of X are equal?

    :meta private:
    """
    X_srt = np.sort(abs(X))
    thresh_ind = int(T * len(X))
    thresh_val = X_srt[-thresh_ind]
    X[abs(X) < thresh_val] = 0
    return X


def _sign(z):
    abs = np.abs(z)
    z[abs == 0] = 0
    abs[abs == 0] = 1
    return z / abs


@contextmanager
def suppress_stdout():
    """
    Suppresses stdout from some healpy functions
    :meta private:
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


def map2alm(image, lmax, **kwargs):
    with suppress_stdout():
        return hp.map2alm(image, lmax, **kwargs)


def alm2map(alm, nside, **kwargs):
    with suppress_stdout():
        return hp.alm2map(alm, nside, **kwargs)


def _multires_bandlimits(L, B, J_min, dirs=1, spin=0):
    phi_l, psi_lm = pys2let.wavelet_tiling(B, L, dirs, J_min, spin)
    psi_l = np.zeros((psi_lm.shape[1], L), dtype=complex)
    for j, psi in enumerate(psi_lm.T):
        psi_l[j, :] = np.array([psi[el ** 2 + el] for el in range(L)])
    gamma_l = np.vstack([phi_l, psi_l])
    bandlimits = np.zeros(gamma_l.shape[0], dtype=int)
    for j, gamma in enumerate(gamma_l):
        bandlimits[j] = np.nonzero(gamma)[0].max() + 1
    return bandlimits


def chebyshev1(X, order):
    """
    Calculates the Chebyshev polynomial of the first kind of the given order at point X.
    Uses the recurrence relation

    :math:`T_{k+1}(X) = 2XT_{k}(X) - T_{k-1}(X)`

    :math:`T_{1}(X) = X`

    :math:`T_{0}(X) = 1`

    :param X: point at which to calulate :math:`T_{k+1}`
    :param int order: polynomial order

    :return: value of Chebyshev polynomial of the first kind of the given order
    """
    if order < 0:
        raise ValueError("order must be >= 0")
    elif order == 0:
        return 1
    elif order == 1:
        return X
    else:
        return 2 * X * chebyshev1(X, order - 1) - chebyshev1(X, order - 2)


def chebyshev2(X, order):
    """
    Calculates the Chebyshev polynomial of the second kind of the given order at point X.
    Uses the recurrence relation

    :math:`U_{k+1}(X) = 2XU_{k}(X) - U_{k-1}(X)`

    :math:`U_{1}(X) = 2X`

    :math:`U_{0}(X) = 1`

    :param X: point at which to calulate :math:`U_{k+1}`
    :param int order: polynomial order

    :return: value of Chebyshev polynomial of the second kind of the given order
    """
    if order < 0:
        raise ValueError("order must be >= 0")
    elif order == 0:
        return 1
    elif order == 1:
        return 2 * X
    else:
        return 2 * X * chebyshev2(X, order - 1) - chebyshev2(X, order - 2)


def cheb1der(X, order):
    """
    Evaluates the derivative of the Chebyshev polynomial of the first kind of the given order at point X.
    Uses the relation

    :math:`dT_{n}/dx = nU_{n-1}`

    :param X: point at which to calulate :math:`dT_{n}/dx`
    :param int order: polynomial order

    :return: derivative of Chebyshev polynomial of the second kind of the given order
    """
    if order < 0:
        raise ValueError("order must be > 0")
    elif order == 0:
        return 0
    else:
        return order * chebyshev2(X, order - 1)


def pixel_area(r, theta1, theta2, phi1, phi2):
    """
    Calculates area of a spherical rectangle.  Angles must be given in radians.

    :params float r: radius
    :params float theta1: starting colatitude
    :params float theta2: ending colatitude
    :params float phi1: starting longitude
    :params float phi2: ending longitude

    :return: area of rectangle in squared radians
    """
    return r ** 2 * (np.cos(theta1) - np.cos(theta2)) * (phi2 - phi1)


def polar_cap_area(r, theta):
    """
    Calculates the area of a polar cap.

    :params float r: radius
    :params float theta: colatitude in radians of bottom of cap

    :return: area of polar cap in squared radians
    """
    return 2 * np.pi * r ** 2 * (1 - np.cos(theta))


def calc_pixel_areas(L, r=1):
    """
    Calculates the areas of all the pixels in `MW <https://arxiv.org/abs/1110.6298>` sampling.

    :param int L: bandlimit
    :param float r: radius

    :return: array of pixel areas with shape :math:`(L, 2L-1)`
    """
    thetas, phis = pyssht.sample_positions(L)
    nthetas, nphis = thetas.shape[0], phis.shape[0]
    areas = np.zeros((nthetas, nphis), dtype=float)
    phis = np.append(phis, [2 * np.pi])
    areas[0] = polar_cap_area(r, thetas[0]) / nphis
    for t, theta1 in enumerate(thetas[:-1]):
        theta2 = thetas[t + 1]
        for p, phi1 in enumerate(phis[:-1]):
            phi2 = phis[p + 1]
            areas[t + 1][p] = pixel_area(r, theta1, theta2, phi1, phi2)
    return areas


def mw_weights(m):
    if m == 1:
        w = 1j * np.pi / 2
    elif m == -1:
        w = -1j * np.pi / 2
    elif m % 2 == 0:
        w = 2.0 / (1.0 - m * m)
    else:
        w = 0

    return w


def weights_theta(L):
    wr = np.zeros(2 * L - 1, dtype=complex)
    for i, m in enumerate(range(-(L - 1), L)):
        wr[i] = mw_weights(m) * np.exp(-1j * m * np.pi / (2 * L - 1))
    wr = (np.fft.fft(np.fft.ifftshift(wr)) * 2 * np.pi / (2 * L - 1) ** 2).real
    return wr


def mw_map_weights(L):
    """
    Calculates exact quadrature weights for `MW <https://arxiv.org/abs/1110.6298>` sampling.

    :param int L: bandlimit

    :return: 1D array of quadrature weights with shape :math:`(L(2L-1),)`
    """
    wr = weights_theta(L)
    q = np.copy(wr[0:L])
    for i, j in enumerate(range(2 * L - 2, L - 1, -1)):
        q[i] = q[i] + wr[j]
    Q = np.outer(q, np.ones(2 * L - 1)).flatten()
    return Q


def s2_integrate(f, L):
    """
    Integrates a spherical image over the whole sphere

    :param array f: spherical image as 1D array of shape :math:`(L(2L-1),)`
    :param int L: bandlimit

    :return: integral of :code:`f` over the sphere

    .. todo::
       Infer :code:`L` from shape of :code:`f`
    """
    f_weighted = mw_map_weights(L) * f
    return f_weighted.sum()


def norm(x):
    return np.linalg.norm(x)


def snr(signal, noise):
    """
    Calculates the signal to noise ratio

    :math:`\mathrm{SNR} = 20\log_{10}\left(\\frac{\|\mathrm{signal}\|_2}{\|\mathrm{noise}\|_2}\\right)`

    :param signal: signal array
    :param noise: noise array

    :return: signal-to-noise ratio in decibels
    """
    return 20 * np.log10(norm(signal) / norm(noise))


def build_mask(L, size=20):
    """"
    Builds a mask for the galactic plane and ecliptic
    0 at positions to be masked
    i.e. to apply mask do map * mask

    size is the width of each mask in degrees

    Mask in MW format
    """
    mask = np.ones(pyssht.sample_shape(L))
    thetas, phis = pyssht.sample_positions(L)
    for i, t in enumerate(thetas):
        for j, p in enumerate(phis):
            if np.abs(90 - np.degrees(t)) < size:
                mask[i, j] = 0

    thetaarray, phiarray = pyssht.sample_positions(L, Grid=True)
    thetaarray = np.degrees(thetaarray) - 90
    phiarray = np.degrees(phiarray) - 180

    c = SkyCoord(phiarray, thetaarray, unit="deg")
    d = c.transform_to("galactic")
    degm = np.abs(d.b.degree)
    for i in range(L):
        for j in range(2 * L - 1):
            if degm[i, j] < size:
                mask[i, j] = 0

    return mask



from pys2let import mw_size
from scipy import sparse
import numpy as np

from pxmcmc.measurements import Identity, PathIntegral
from pxmcmc.transforms import SphericalWaveletTransform


class ForwardOperator:
    """
    Base Forward operator. Combines a transform and a measurement operator.

    :param data: observed data vector
    :param sig_d: observed data error.  Can be a single float, vector or covariance matrix.
    :param string setting: `analysis` or `synthesis`
    :param transform: type :class:`transforms.Transform` to transform between bases
    :param measurement: type :class:`measurements.Measurement` to predict observed data
    :param int nparams: number of sampled parameters/dimensions.  Depends on parameterisation and setting.
    """

    def __init__(
        self, data, sig_d, setting, transform=None, measurement=None, nparams=None
    ):
        self.data = data
        self.invcov = self._build_inverse_covariance_matrix(sig_d)
        if setting not in ["analysis", "synthesis"]:
            raise ValueError
        self.setting = setting
        if transform is not None:
            self.transform = transform
        if measurement is not None:
            self.measurement = measurement
        if nparams is not None:
            self.nparams = nparams

    def forward(self, X):
        """
        Forward modelling, with or without basis transformation depending on :code:`setting`.

        :param X: MCMC sample
        :return: data predictions
        """
        if self.setting == "analysis":
            return self._forward_analysis(X)
        else:
            return self._forward_synthesis(X)

    def calc_gradg(self, preds):
        """
        Calculates gradient of data fidelity.  Assumes Gaussian data errors
        
        :param preds: data predictions of current MCMC sample
        :return: gradient of Gaussian data fidelity
        """
        if self.setting == "analysis":
            return self._gradg_analysis(preds)
        else:
            return self._gradg_synthesis(preds)

    def _forward_analysis(self, X):
        return self.measurement.forward(X)

    def _forward_synthesis(self, X):
        return self.measurement.forward(self.transform.inverse(X))

    def _gradg_analysis(self, preds):
        return self.measurement.adjoint(
            (self.invcov @ sparse.csr_matrix(preds - self.data).T).toarray().flatten()
        )

    def _gradg_synthesis(self, preds):
        return self.transform.inverse_adjoint(self._gradg_analysis(preds))

    def _build_inverse_covariance_matrix(self, sig_d):
        if isinstance(sig_d, np.ndarray) and len(sig_d.shape) == 2:
            if sig_d.shape[0] != sig_d.shape[1]:
                raise ValueError("Covariance matrix should be square")
            return sparse.linalg.inv(sig_d)

        var = sig_d ** 2
        if np.iscomplexobj(self.data) and not np.iscomplexobj(var):
            var = var / np.sqrt(2) * (1 + 1j)
        if isinstance(var, float) or isinstance(var, int) or isinstance(var, complex):
            return sparse.identity(len(self.data)).dot(1 / var)
        elif var.size == len(self.data) and len(var.shape) == 1:
            return sparse.diags(1 / var)
        else:
            raise TypeError("sig_d must be a float scalar, vector or 2D matrix")


class SphericalWaveletTransformOperator(ForwardOperator):
    """
    Forward operator with a spherical wavelet transform and identity operator.

    :param int L: angular bandlimit
    :param float B: wavelet scale parameter
    :param int J_min: minimum wavelet scale
    :param int dirs: azimuthal bandlimit for directional wavelets
    :param int spin: spin number of spherical signal
    """
    def __init__(self, data, sig_d, setting, L, B, J_min, dirs=1, spin=0):
        transform = SphericalWaveletTransform(
            L,
            B,
            J_min,
            dirs=dirs,
            spin=spin,
        )
        measurement = Identity(len(data), mw_size(L))

        if setting == "analysis":
            nparams = mw_size(L)
        else:
            nparams = transform.ncoefs

        super().__init__(
            data,
            sig_d,
            setting,
            transform=transform,
            measurement=measurement,
            nparams=nparams,
        )


class PathIntegralOperator(ForwardOperator):
    """
    Forward operator with a spherical wavelet transform and a path integral measurement operator.

    .. todo::
       Since the measurement operator is just a matrix multiplication, can be renamed to something more generic.

    :param array pathmatrix: matrix describing a set of paths on the sphere
    :param int L: angular bandlimit
    :param float B: wavelet scale parameter
    :param int J_min: minimum wavelet scale
    :param int dirs: azimuthal bandlimit for directional wavelets
    :param int spin: spin number of spherical signal
    """
    def __init__(self, pathmatrix, data, sig_d, setting, L, B, J_min, dirs=1, spin=0):
        transform = SphericalWaveletTransform(
            L,
            B,
            J_min,
            dirs=dirs,
            spin=spin,
        )
        measurement = PathIntegral(pathmatrix)

        if setting == "analysis":
            nparams = mw_size(L)
        else:
            nparams = transform.ncoefs

        super().__init__(
            data,
            sig_d,
            setting,
            transform=transform,
            measurement=measurement,
            nparams=nparams,
        )
